!+

! Module: omfastflat_mod
!
! Authors:  Jon Rainger + Tim Harries - MSSL
!          
! Description: Flatfields fast mode images
! 
! Comments: 
! 
! $Id: module.f90,v 1.5 1998/01/13 08:39:15 gvacanti Exp $
!- 
! History:
!          1999 Jan: updated psw - spr 553
!          2001 May 08 (V.Y.): modificated slewFlat array correction for the flat field 
!                              (eliminated windowX0P(0) and windowY0P(0) from the indexes
!                               because rawXP and rawYP alreay containes the origin of the window)
!          2001 May 08 (V.Y.)  filling in the fastImgP array in the correct way 
!          2001 October 17 (VNY) - data set in the /test directory has been updated
!          2001 October 19 (VNY) - more robust call to CheckArrayBounds subroutine
!          2001 October 22 (VNY) - a claryfying message has been added to CheckArrayBounds subroutine
!          2001 October 23 (VNY) - additional checks and integer conversions to CheckArrayBounds
!                               to prevent a falure at Cosine3 machine
!          2001 October 24 (VNY) - pointers to window parameters (windowX0P, windowY0P, windowDXP, windowDYP)
!                                replaced with int32 variables in the cases when these parameters are 
!                                passed to inner subroutines (in order to avoid possible subscripts)
!          2001 October 25 (VNY) a bug with the use of factorP pointer has been fixed to rectify 
!                                possible test-harness in omdetect task 
!          2002 January 08 (VNY) DEPEND file has been updated to use higher version of ssclib (2.0)
!          2007 June 06 (VNY) Protecting from crashing in the case of dealing with the event list
!                             file, which has been already once processed by omfastflat (fixing
!                             the SSC-SPR-3667)
!          2007 June 07 (VNY) Introducing some analysis of the image as to the possible modulo-8
!                             corruption and setting a new flag MOD8FLAG to 1 to indicate  
!                             the situation when the image is possible corrupted by coincidence 
!                             losses combined with the modulo-8 pattern. For the normal image
!                             the flag is set to 0
!          2009 Oct 23 (VNY) Revisiting the criterium for MOD8FLAG calculation
!          2010 Jan 26 (VNY) Issuing a message indicating the mod8flag value
!          2010 March 11 (VNY) Fixing a bug related to the usage of pointers, which 
!                              was affecting the calculation of the mod-8 modulation factor
!                              (ssc-spr-6599). Increasing the threshold for flagging the 
!                              exposure as been affected by the mod-8 noise from 0.25 to 0.35 
!                              to avoid possible sudden increases of the source extraction 
!                              radius for moderately mod-8 corrupted exposures. 
! 2010-03-24 (VNY) Improved the calculation of the mod-8 modulation factor, which was 
!                  overestimated for some of the Fast exposures due to the source size
!                  and position resemblint the mod-8 pattern
! 2011-09-29 (VNY) Improved the calculation of the mod-8 modulation factor by removing
! the first and the last raws and columns from the analysis (as these raws and columns are 
! likely to be affected by the tracking shifts)
! 2011-10-04 (VNY) Further improvements in the calculation of the mod8 modulation factor
! by excluding the source region from the analysis
!
module omfastflat_mod

  use dal
  use cal
  use param
  use errorHandling
  use copyTable_mod
  use sasdal

  implicit none

  private

  public omfastflat

contains

  !---------------------------------------------------------------------
  subroutine omfastflat
    !---------------------------------------------------------------------

    character(len=255) :: set                                         ! OM Fast Mode Event List filename (output of OMFASTSHIFT).
    character(len=255) :: slewFlatSet                                 ! OM Inorbit Flatfield filename (output of OMFLATGEN).
    character(len=255) :: fastImgSet                                  ! Flatfielded OM Fast Mode Image filename.
    character(len=255) :: OSWFlatSet                                  ! Data product file: OM OSW flat field filename. 

    type(DataSetT) :: inDataSetH                                      ! Input dataset handle.
    type(DataSetT) :: slewFlatDataSetH                                ! OM Inorbit Flatfield dataset handle.
    type(DataSetT) :: fastImgDataSetH                                 ! OM Fast Mode Image dataset handle.
    type(DataSetT) :: OSWFlatDataSetH                                 ! OM OSW flat field dataset handle.

    type(ArrayT) :: slewFlatArrayH                                    ! Input array handle.
    type(ArrayT) :: fastImgArrayH                                     ! Output array handle.
    type(ArrayT) :: qualityArrayH                                     ! Output array handle.
    type(ArrayT) :: OSWFlatArrayH                                     ! Output array handle.
    type(TableT) :: inTableH                                          ! Input table handle.
    type(TableT) :: modesTableH                                       ! Modes table dataset handle.

    type(ColumnT), dimension(5) :: inColumnH                          ! Input column handle.
    integer, parameter :: noColumns = 20 !+
    integer(kind=int8), dimension(noColumns) :: outColumnFlag          ! flags for checking the column existence
    type(ColumnT), dimension(4) :: modesColumnH                       ! Modes column handle.
    type(ColumnT), dimension(2) :: outColumnH                         ! Output column handle.

    integer(kind=int16), dimension(:), pointer :: rawXP               ! Raw x column pointer.
    integer(kind=int16), dimension(:), pointer :: rawYP               ! Raw y column pointer.
    integer(kind=int16), dimension(:), pointer :: corrXP              ! Raw x column pointer.
    integer(kind=int16), dimension(:), pointer :: corrYP              ! Raw y column pointer.
    integer(kind=int8), dimension(:), pointer :: winFlagP             ! WinFlag pointer
    integer(kind=int32), dimension(:), pointer :: windowX0P           ! WindowX0 pointer.
    integer(kind=int32), dimension(:), pointer :: windowY0P           ! WindowY0 pointer.
    integer(kind=int32), dimension(:), pointer :: windowDXP           ! WindowDX pointer.
    integer(kind=int32), dimension(:), pointer :: windowDYP           ! WindowDY pointer.
!---VNY 24.10.2001
    integer(kind=int32) :: windowX0, windowY0, windowDX, windowDY

    real(kind=single), dimension(:, :), allocatable :: slewFlat       ! Real array pointer.
    real(kind=single), dimension(:, :), pointer :: slewFlatP          ! Real array pointer.

    integer(kind=int16), dimension(:, :), pointer :: badPixMapP       ! Integer array pointer to bad pixel map.   

    integer(kind=int8), dimension(:), pointer :: badPixFlagP          ! Bad pixel flag column pointer.
    real(kind=single), dimension(:), pointer :: factorP               ! QE Factor column pointer.

    integer(kind=int8), dimension(:, :), pointer :: qualityP          ! Quality array pointer.
    real(kind=single), dimension(:, :), pointer :: fastImgP           ! Real array pointer.

    real(kind=single), dimension(:, :), pointer :: OSWFlatImgP        ! OSW flat field array pointer.

    integer(kind=int32), dimension(2) :: imageDim                     ! Image dimension.
    integer(kind=int32) :: nEvents                                    ! Number of events.
    integer(kind=int32) :: i, i1                                          ! Counter.
    integer(kind=int32) :: j, j1                                          ! Counter.
!----------------Added by VY 8.05.2001
    integer(kind=int32) :: indexX                                     ! for calculation of the output image array index  
    integer(kind=int32) :: indexY
    integer :: intRawX, intRawY 
!------------------
    character(len=255) :: dataMode                                    ! Input file data mode.
    character(len=255) :: binBPE                                      ! Input file BPE binning keyword.
    character(len=255) :: filterMode                                  ! Input file filter mode
    real(kind=single) :: imgMax, imgMin, imgAver, nAver
    real(kind=single) :: belowAver, aboveAver, nBelow, nAbove
    real(kind=single) :: nZeroPixels, zeroFactor
    real(kind=single) :: exposureTime
    real(kind=single) :: averRate, maxRate

!***********************************************
!    real(kind=single) :: vsum, nvsum
!    !--------------------------------------------------------
!    !! Following are the variables for saving testimage77.fit
!    character(len=255) :: outSetCheck
!    integer(int32) :: outDimCheck(2) 
!    type(DataSetT) :: outDataSetCheckH 
!    type(ArrayT) :: outArrayCheckH 
!    real(single), dimension(:, :), pointer :: outOSWcheckP  
!    !--------------------------------------------------------
!    real(kind=single), pointer :: imImage2(:,:)
!    real(kind=single), pointer :: imImage3(:,:)
!    real(kind=single), pointer :: imImage4(:,:)
!    real(kind=single), pointer :: sliceX(:), sliceX2(:)
!    real(kind=single) :: vMax, vMin, vMaxmin, vMin0, vMaxmin0, maxMinDiff, inversedWeight
!    integer :: nvMax, nvMin

    real(kind=single) :: mod8xfactor, mod8yfactor, mod8factor, mod8factorMin
    integer(kind=int32) :: dimtst(2),nxTst, nyTst
    !real(single), allocatable, dimension(:, :) :: checkOSW 
    !integer(int32) :: allocErr  

    mod8xfactor=0.0
    mod8yfactor=0.0
    mod8factor=0.0
    mod8factorMin=0.0

    ! get task parameters.

    set = stringParameter("set")
    slewFlatSet = stringParameter("slewflatset")
    fastImgSet = stringParameter("fastimgset")
    OSWFlatSet = stringParameter("oswflatset")


    ! Check data types of ODF input files

    ! **** JFR **** need to check if input is an Inorbit flatfield.

    ! Get a hold on the Inorbit flat field File.

    
    slewFlatDataSetH = dataSet(slewFlatSet, read)

    slewFlatArrayH = array(slewFlatDataSetH, "PRIMARY", Read)
    slewFlatP => real32Array2Data(slewFlatArrayH)


    ! Get a hold on the ODF OM Fast Mode Event List File.

    inDataSetH = dataSet(set, modify)

    ! Determine file type from file name in primary header.


    dataMode = trim(stringAttribute(inDataSetH, "DATAMODE")) 

    ! Determine filter type from the input data
    
    filterMode = trim(stringAttribute(inDataSetH, "FILTER"))

    exposureTime = real32Attribute(inDataSetH, "EXPOSURE")

    ! Determine if BPE binning is on/off

    binBPE = trim(stringAttribute(inDataSetH, "BINBPE")) 

    if (dataMode == "FAST") then

       ! Get a handle on the tracking history data binary table extension.

       inTableH = table(inDataSetH, "OMFAE1")

       ! Get handles and pointers on the event list binary table extension columns.

       inColumnH(1) = column(inTableH, "RAWX", read)
       rawXP => int16Data(inColumnH(1))

       inColumnH(2) = column(inTableH, "RAWY", read)
       rawYP => int16Data(inColumnH(2))

       inColumnH(3) = column(inTableH, "CORR_X", read)
       corrXP => int16Data(inColumnH(3))

       inColumnH(4) = column(inTableH, "CORR_Y", read)
       corrYP => int16Data(inColumnH(4))

       inColumnH(5) = column(inTableH, "WIN_FLAG", Read)
       winFlagP => int8data(inColumnH(5))


       ! Get the number of rows in the event list binary table extension.

       nEvents = numberOfRows(inTableH)

       ! Get window information from modes extension.

       modesTableH = table(inDataSetH, "MODES")

       modesColumnH(1) = column(modesTableH, "WINDOWX0", read)
       windowX0P => int32data(modesColumnH(1))

       modesColumnH(2) = column(modesTableH, "WINDOWY0", read)
       windowY0P => int32data(modesColumnH(2))

       modesColumnH(3) = column(modesTableH, "WINDOWDX", read)
       windowDXP => int32data(modesColumnH(3))

       modesColumnH(4) = column(modesTableH, "WINDOWDY", read)
       windowDYP => int32data(modesColumnH(4))

!---VNY 24.10.2001
       windowX0 = windowX0P(0)
       windowY0 = windowY0P(0)
       windowDX = windowDXP(0)
       windowDY = windowDYP(0)
       
    else

       ! Incompatible file.

       call fatal("badFileMode","Value for keyword DATAMODE is incompatible")

    end if


    ! Set the state of the cal.

    call CAL_setState(OM, 1, 0)

    ! Get the bad pixel map.

    call CAL_getBadPixelMap(badPixMapP)
    outColumnFlag=1
    ! Add factor and bad pixel columns to event list file.
    if (hasColumn(inTableH, "FACTOR")) then
       outColumnH(1) = column(inTableH, "FACTOR", modify)
       factorP => real32data(outColumnH(1))
    else
       outColumnFlag(1)=0
       outColumnH(1) = addColumn(inTableH, "FACTOR", real32)
       factorP => real32data(outColumnH(1))
    endif
    if (hasColumn(inTableH, "BADPIX")) then
       outColumnH(2) = column(inTableH, "BADPIX", modify)
       badPixFlagP => int8data(outColumnH(2))
    else
      outColumnFlag(2)=0
      outColumnH(2) = addColumn(inTableH, "BADPIX", integer8)
      badPixFlagP => int8data(outColumnH(2))
    endif

    ! Initialise pointer arrays to zero.

    factorP = 0.0
    badPixFlagP = 0

    ! Calculate factor and bad pixel flag.

    ! If BPE binning off then need to rebin flatfield.


    if (binBPE == "T") then
       allocate(slewFlat(0:1023,0:1023))
       slewFlat = slewFlatP
    else 

       allocate(slewFlat(0:2047,0:2047))       
       do i= 1, 1024
          do j = 1, 1024
!--- VNY 23.10.2001 Additional conversions for Cosine3 machine
             intRawX=2*i-2
             intRawY=2*j-2
             call CheckArrayBounds(slewFlat, intRawX,intRawY )
             intRawX=2*i-1
             intRawY=2*j-1
             call CheckArrayBounds(slewFlat, intRawX,intRawY)
             intRawX=2*i-2
             intRawY=2*j-1
             call CheckArrayBounds(slewFlat, intRawX, intRawY)
             intRawX=2*i-1
             intRawY=2*j-2
             call CheckArrayBounds(slewFlat, intRawX, intRawY)

             slewFlat(2*i-2,2*j-2) = slewflatP(i-1,j-1)
             slewFlat(2*i-1,2*j-1) = slewflatP(i-1,j-1)
             slewFlat(2*i-2,2*j-1) = slewflatP(i-1,j-1)
             slewFlat(2*i-1,2*j-2) = slewflatP(i-1,j-1)
          end do
       end do
    end if


    do i = 0, nEvents - 1
       intRawX=rawXP(i)
       intRawY=rawYP(i)
!---VNY 23.10.2001 Additional check for Cosine3 machine
       if (intRawX .lt. 2048 .and. intRawY .lt. 2048) then
          call CheckArrayBounds(slewFlat, intRawX, intRawY)
       else
          if (intRawX .ge. 2048) then
             call warning("X exceeds 2047") 
             intRawX=2047
           end if
          if (intRawY .ge. 2048) then
             call warning("Y exceeds 2047") 
             intRawY=2047
           end if
       end if

       if(slewFlat(rawXP(i), rawYP(i)) /= 0) then
          factorP(i) = (slewFlat(rawXP(i), rawYP(i)))**(-1.0)
       else
          factorP(i) = 0
       end if

       intRawX=rawXP(i)
       intRawY=rawYP(i)
       call CheckInt16ArrayBounds(badPixMapP, intRawX, intRawY)

       if (badPixMapP(rawXP(i), rawYP(i)) /= 0) then
          badPixFlagP(i) = 1
       end if
    end do


    ! Create flatfielded fast mode image file.
 
    fastImgDataSetH = dataSet(fastImgSet, Create)

    ! Copy attributes.

    call copyAllAttributes(attributable(inDataSetH), attributable(fastImgDataSetH))

    !   call setAttribute(outDataSetH, "CONTENT", "OM OSW IMAGE", "Content of file")
    !   call setAttribute(OutDataSetH, "HDUCLASS", "OGIP", "Format conforms to OGIP/GSFC conventions")
    !   call setAttribute(OutDataSetH, "HDUCLAS1", "IMAGE", "File contains an image")
    !   call setAttribute(OutDataSetH, "HDUCLAS2", "TOTAL", "Total counts")

    imageDim(1) = windowDX
    imageDim(2) = windowDY
    !write(*,*) "L365 imageDim(1)=", imageDim(1), " imageDim(2)=", imageDim(2)
    ! Get image handle and pointer.

    fastImgArrayH = addArray(fastImgDataSetH, "PRIMARY", real32, imageDim)
    fastImgP => real32array2data(fastImgArrayH)

    dimTst = shape(fastImgP)
    nxTst = dimTst(1)
    nyTst = dimTst(2) 
    !write(*,*) "L375 nxTst=", nxTst, " nyTst=", nyTst

    ! Get quality array handle and pointer.

    qualityArrayH = addArray(fastImgDataSetH, "QUALITY", integer8, imageDim)
    qualityP => int8array2data(qualityArrayH)

    dimTst = shape(qualityP)
    nxTst = dimTst(1)
    nyTst = dimTst(2) 
    !write(*,*) "L385 nxTst=", nxTst, " nyTst=", nyTst

    !allocate (checkOSW(0: nxTst-1, 0: nyTst-1), stat = allocErr)

    ! Build image and quality array.

    !fastImgP = 0.0
    do i=0,windowDX - 1
       do j=0,windowDY - 1
          fastImgP(i,j)=0.0
          !checkOSW(i,j)=0.0
       enddo
    enddo


!****************** 2009-10-23 *******************************
!    allocate (imImage2(0 : windowDX-1, 0 : windowDY-1))
!    allocate (imImage3(0 : windowDX-1, 0 : windowDY-1))
!    allocate (imImage4(0 : windowDX-1, 0 : windowDY-1))
!    allocate (sliceX(0 : windowDX-1))
!    allocate (sliceX2(0 : windowDY-1))
!-------------------------Corrections by V.Y. 8.05.2001 (filling in the fastImgP array in a correct way)
    !qualityP = 0
    do i=0,windowDX - 1
       do j=0,windowDY - 1
          qualityP(i,j)=0
       enddo
    enddo

    do i = 0 , nEvents - 1

       !if (i .eq. 0) write(*,*) "L390 winFlagP(0)=", winFlagP(i)
       !write(*,*) "L391 i=", i, " winFlagP=", winFlagP(i)

       if (winFlagP(i) .ne. 1) then          
          indexX=corrXP(i)-windowX0
          ! 2010-03-11: if the coordinates are out of the window, discard this event
          !if (indexX .lt. 0) then
          !   indexX=0
          !end if

          if (indexX .gt. imageDim(1)) then
             indexX=-indexX
          end if          

          indexY=corrYP(i)-windowY0

          !if (indexY .lt. 0) then
          !   indexY=0
          !end if

          if (indexY .gt. imageDim(2)) then
             indexY=-indexY
          end if

          if (indexX .ge. 0 .and. indexX .lt. imageDim(1) .and. indexY .ge. 0 .and. indexY .lt. imageDim(2)) then
             fastImgP(indexX, indexY) = fastImgP(indexX, indexY) + factorP(i) 
             if (badPixFlagP(i) == 1) then
                qualityP(indexX, indexY) = 1
             end if
          end if
       end if
    end do

! Analysis of the image
    imgMax=0.0
    imgMin=1.0E+10
    imgAver=0.0
    do i=0,windowDX - 1
       do j=0,windowDY - 1
          imgAver=imgAver+fastImgP(i,j)
          nAver=nAver+1.0
          if (fastImgP(i,j) .gt. imgMax) imgMax=fastImgP(i,j)
          if (fastImgP(i,j) .lt. imgMin) imgMin=fastImgP(i,j)
       enddo
    enddo

    if (nAver .gt. 1.0) imgAver=imgAver/nAver
    !write(*,*) "L473 imgAver=", imgAver, " nAver=", nAver

! computing the averages below and above the threshold
    belowAver=0.0
    nBelow=0.0
    aboveAver=0.0
    nAbove=0.0
    nZeroPixels=0
    do i=0,windowDX - 1
       do j=0,windowDY - 1
          if (fastImgP(i,j) .ge. imgAver/10.0) then
             aboveAver=aboveAver+fastImgP(i,j)
             nAbove=nAbove+1.0
          else
             belowAver=belowAver+fastImgP(i,j)
             nBelow=nBelow+1.0
          endif
          if (fastImgP(i,j) .lt. 0.01) nZeroPixels=nZeroPixels+1.0
       enddo
    enddo
    !write(*,*) "L494 imgAver=", imgAver, " nAbove=", nAbove, " nBelow=", nBelow
    
    if (nAbove .gt. 1.0) aboveAver=aboveAver/nAbove
    if (nBelow .gt. 1.0) belowAver=belowAver/nBelow
    zeroFactor=0.0
    if (nAver .gt. 0.0) zeroFactor=nZeroPixels/nAver
    averRate=aboveAver
    maxRate=imgMax
    if (exposureTime .gt. 0.01) then
       averRate=aboveAver/exposureTime
       maxRate=imgMax/exposureTime
    endif

    !write(*,*) "L507 aboveAver=", aboveAver, " belowAver=", belowAver, " zeroFactor=", zeroFactor
    !write(*,*) "L508 averRate=", averRate, " maxRate=", maxRate

!*********************************************************
! 2009-10-23 Revisiting the criterium for mod-8 noise flag
!*********************************************************
! computing the averages below and above the threshold

    call CheckMod8(fastImgP,mod8xfactor,mod8yfactor)
    !call CheckMod8(checkOSW,mod8xfactor,mod8yfactor)
    !write(*,*) "L522 mod8xfactor=", mod8xfactor, " mod8yfactor=", mod8yfactor
    mod8factor=max(mod8xfactor, mod8yfactor)
    mod8factorMin=min(mod8xfactor, mod8yfactor)

    write(errstr,'(a,f5.2)') "Mod8-noise modulation factor =", mod8factor
    call message(AppMsg, 2)

    !write(*,*) "L468 mod8xfactor=", mod8xfactor, " mod8yfactor=", mod8yfactor

    !do i=0,windowDX - 1
    !   do j=0,windowDY - 1
    !      fastImgP(i,j)=checkOSW(i,j)
    !   enddo
    !enddo
    !deallocate(checkOSW)

!********************************************************
    !write(*,*) "L470 maxRate=", maxRate, " zeroFactor=", zeroFactor, " mod8Factor=", mod8factor
    if (maxRate .gt. 10.0 .and. zeroFactor .gt. 0.1) then
       call warning("Modulo-8 noise", "the source is very bright and is likely to be corrupted") 
       call setAttribute(fastImgDataSetH, "MOD8FLAG", 1, "Flag for modulo-8 corruption (0 is ok)")
       call setAttribute(inDataSetH, "MOD8FLAG", 1, "Flag for modulo-8 corruption (0 is ok)")
       call setAttribute(fastImgDataSetH, "MOD8FACT", mod8factor, "Mod-8 noise factor")
       call setAttribute(inDataSetH, "MOD8FACT", mod8factor, "Mod-8 noise factor")
    else
       !* 2010-03-11 Increasing the criterium for the mod-8 corruption flag 
       !if (mod8factor .gt. 0.25) then
       if (mod8factor .gt. 0.35 .and. mod8factorMin .gt. 0.1) then
          call warning("Modulo-8 noise", "the source is very bright and likely to be corrupted") 
          call setAttribute(fastImgDataSetH, "MOD8FLAG", 1, "Flag for modulo-8 corruption (0 is ok)")
          call setAttribute(inDataSetH, "MOD8FLAG", 1, "Flag for modulo-8 corruption (0 is ok)")
          call setAttribute(fastImgDataSetH, "MOD8FACT", mod8factor, "Mod-8 noise factor")
          call setAttribute(inDataSetH, "MOD8FACT", mod8factor, "Mod-8 noise factor")
       else
          call setAttribute(fastImgDataSetH, "MOD8FLAG", 0, "Flag for modulo-8 corruption (0 is ok)")
          call setAttribute(inDataSetH, "MOD8FLAG", 0, "Flag for modulo-8 corruption (0 is ok)")
          call setAttribute(fastImgDataSetH, "MOD8FACT", mod8factor, "Mod-8 noise factor")
          call setAttribute(inDataSetH, "MOD8FACT", mod8factor, "Mod-8 noise factor")
       endif
    endif
!---VNY 25.10.2001
    ! Release the cal memory.
    call cal_releaseMemory(badPixMapP)
    ! Release handles on data.
    call release(outColumnH(1))
    call release(outColumnH(2))

    call copyTable(inDataSetH, fastImgDataSetH, "MODES")

    ! Create OM OSW flat field data product.

    OSWFlatDataSetH = dataSet(OSWFlatSet, create)

    call copyAllAttributes(attributable(inDataSetH), attributable(OSWFlatDataSetH))

    call setAttribute(OSWFlatDataSetH, "CONTENT", "OM OSW FLAT FIELD", "Content of file")
    call setAttribute(OSWFlatDataSetH, "HDUCLASS", "OGIP", "Format conforms to OGIP/GSFC conventions")
    call setAttribute(OSWFlatDataSetH, "HDUCLAS1", "IMAGE", "File contains an image")
    call setAttribute(OSWFlatDataSetH, "HDUCLAS2", "DETMAP", "Detector map")

    OSWFlatArrayH = addArray(OSWFlatDataSetH, "PRIMARY", real32, imageDim)
    OSWFlatImgP => real32array2data(OSWFlatArrayH)

    do i=0,windowDX - 1
       do j=0,windowDY - 1
          intRawX=windowX0+i
          intRawY=windowY0+j

!---VNY 23.10.2001 Additional check for Cosine3 machine
          if (intRawX .lt. 2048 .and. intRawY .lt.2048) then
             call CheckArrayBounds(slewFlat, intRawX, intRawY)
          else
             if (intRawX .ge. 2048) then
                call warning("X exceeds 2047")
                intRawX=2047
             end if
             if (intRawY .ge.2048) then
                call warning("Y exceeds 2047")
                intRawY=2047
             end if
          end if
!---VNY 23.10.2001 Additional check for Cosine3 machine
          i1=i
          j1=j
          if (i1 .lt. 2048 .and. j1 .lt. 2048) then 
             call CheckArrayBounds(OSWFlatImgP, i1,j1)
          else
             if (i1 .ge. 2048) then
                call warning("I exceeds 2047")
                i1=2047
             end if
             if (j1 .ge. 2048) then
                call warning("J exceeds 2047")
                j1=2047
             end if
          end if
          OSWFlatImgP(i1,j1) = slewFlat(intRawX, intRawY)
       end do
    end do

    call copyTable(inDataSetH, OSWFlatDataSetH, "MODES")


    ! Free up memory and pointers.

    call release(modesColumnH(1))
    call release(modesColumnH(2))
    call release(modesColumnH(3))
    call release(modesColumnH(4))
    call release(modesTableH)

    call release(OSWFlatArrayH)
    call release(OSWFlatDataSetH)

    call release(qualityArrayH)
    call release(fastImgArrayH)
    call release(fastImgDataSetH)

    call release(inColumnH(1))
    call release(inColumnH(2))
    call release(inColumnH(3))
    call release(inColumnH(4))
    call release(inColumnH(5))
    call release(inTableH)

    call release(inDataSetH)
    call release(slewFlatArrayH)
    call release(slewFlatDataSetH)


  end subroutine omfastflat


subroutine CheckMod8(image, mod8xfactor, mod8yfactor)
    real(kind=single), intent(in), dimension(: , :) :: image
    real, intent(out) :: mod8xfactor, mod8yfactor 
    integer :: windowDX, windowDY
    integer :: i,j,i1,j1
    real(kind=single) :: imgAver, nAver
    real(kind=single) :: vsum, nvsum
    !--------------------------------------------------------
    !! Following are the variables for saving testimage77.fit
    character(len=255) :: outSetCheck
    integer(int32) :: outDimCheck(2) 
    type(DataSetT) :: outDataSetCheckH 
    type(ArrayT) :: outArrayCheckH 
    real(single), dimension(:, :), pointer :: outOSWcheckP  
    !--------------------------------------------------------
    real(kind=single), pointer :: imImage1(:,:)
    real(kind=single), pointer :: imImage2(:,:)
    real(kind=single), pointer :: imImage3(:,:)
    real(kind=single), pointer :: imImage4(:,:)
    real(kind=single), pointer :: sliceX(:), sliceX2(:)
    real(kind=single) :: vMax, vMin, vMaxmin, vMin0, vMaxmin0, maxMinDiff, inversedWeight
    integer :: nvMax, nvMin, dimax
    integer :: xMax, yMax
    real(kind=single) :: imgMax, threshold
    real(kind=single) :: xsrc, ysrc, xpix,ypix, xweightx, weightx, yweighty, weighty
    real(kind=single) :: dist

    mod8xfactor=0.0
    mod8yfactor=0.0

    windowDX = size(image, 1) 
    windowDY = size(image, 2) 
    dimax=max(windowDX,windowDY)
    !write(*,*) "L677 windowDX=", windowDX, " windowDY=", windowDY

    allocate (imImage1(0 : windowDX-1, 0 : windowDY-1))
    allocate (imImage2(0 : windowDX-1, 0 : windowDY-1))
    allocate (imImage3(0 : windowDX-1, 0 : windowDY-1))
    allocate (imImage4(0 : windowDX-1, 0 : windowDY-1))
    allocate (sliceX(0 : dimax))
    allocate (sliceX2(0 : dimax))

    ! compute the average
    imgMax=-99999.0
    xMax=-1
    yMax=-1
    do i=1,windowDX-1 
       do j=1,windowDY-1
          imgAver=imgAver+image(i,j)
          nAver=nAver+1.0
          if (image(i,j) .gt. imgMax) then
             imgMax=image(i,j)
             xMax=i
             yMax=j
          endif
       enddo
    enddo
    if (nAver .gt. 1.0) imgAver=imgAver/nAver
    !write(*,*) "L702 imgAver=", imgAver, " imgMax=", imgMax, " xMax=", xMax, " yMax=", yMax

    threshold=imgAver+(imgMax-imgAver)*0.7
    !write(*,*) "L707 threshold=", threshold

    
   ! Limit the image by imgAver
    xsrc=-1.0
    ysrc=-1.0
    xweightx=0.0
    yweighty=0.0
    weightx=0.0
    weighty=0.0
    do i=0,windowDX - 1
       do j=0,windowDY - 1
          imImage1(i,j)=image(i+1,j+1)
          ! Compute the bright source position
          if (imImage1(i,j) .gt. threshold) then
             xpix=real(i)
             ypix=real(j)
             xweightx=xweightx+xpix*imImage1(i,j)
             weightx=weightx+imImage1(i,j)
             yweighty=yweighty+ypix*imImage1(i,j)
             weighty=weighty+imImage1(i,j)
          endif
          if (imImage1(i,j) .gt. imgAver) imImage1(i,j)=imgAver
       enddo
    enddo
    if (weightx .gt. 1.0) xsrc=xweightx/weightx
    if (weighty .gt. 1.0) ysrc=yweighty/weighty
    !write(*,*) "L733 xdrc=", xsrc, " ysrc=", ysrc
    
    !----------------------
    !write(*,*) "L715 creating testimage77.fit"
    !outSetCheck="testimage77.fit"
    !outDimCheck(1)=windowDX
    !outDimCheck(2)=windowDY
    !outDataSetCheckH = dataSet(trim(outSetCheck), Create)
    !outArrayCheckH=addArray(outDataSetCheckH, "PRIMARY", &
    !     real32, outDimCheck, "Counts", "IMAGE")!             
    !outOSWcheckP => real32Array2Data(outArrayCheckH)
    !outOSWcheckP = imImage1 
    !! Release handles on data.
    !call release(outArrayCheckH)        
    !call release(outDataSetCheckH) 
    ! 
    ! write(*,*) "L728 Test image 77 saved:", trim(outSetCheck)!
    !--------------------------------

    ! Subtract the cosecutive columns from one another
    imImage2=0.0
    ! left direction
    do i=0,windowDX - 2
       do j=0,windowDY - 1
          imImage2(i,j)=imImage1(i,j)-imImage1(i+1,j)
       enddo
    enddo

    imImage4=0.0
    ! right direction
    do i=1,windowDX - 1
       do j=0,windowDY - 1
          imImage4(i,j)=imImage1(i,j)-imImage1(i-1,j)
       enddo
    enddo

    ! sum 
    ! and also remove all of the pixels within 6-pixel radius around the source
    do i=0,windowDX - 1
       do j=0,windowDY - 1
          imImage2(i,j)=imImage2(i,j)+imImage4(i,j)
          ! remove all of the pixels within 6-pixel radius around the source
          if (xsrc .gt. 0.0 .and. ysrc .gt. 0.0) then
             xpix=real(i)
             ypix=real(j)
             dist=(xpix-xsrc)*(xpix-xsrc)+(ypix-ysrc)*(ypix-ysrc)
             dist=sqrt(dist)
             if (dist .le. 6.0) imImage2(i,j)=-1.0!0.0
          endif
          !***************************************************
          ! 2011-09-29 nullify the extreme raws and columns
          !***************************************************
          if (i .eq. 0 .or. i .eq. windowDX-1 .or. j .eq. 0 .or. j .eq. windowDY-1) then
             imImage2(i,j)=-1.0!0.0
          endif

       enddo
    enddo


      

    !----------------------
    !write(*,*) "L730 creating testimage77.fit"
    !outSetCheck="testimage77.fit"
    !outDimCheck(1)=windowDX
    !outDimCheck(2)=windowDY
    !outDataSetCheckH = dataSet(trim(outSetCheck), Create)
    !outArrayCheckH=addArray(outDataSetCheckH, "PRIMARY", &
    !     real32, outDimCheck, "Counts", "IMAGE")!             
    !outOSWcheckP => real32Array2Data(outArrayCheckH)
    !outOSWcheckP = imImage2 
    !! Release handles on data.
    !call release(outArrayCheckH)        
    !call release(outDataSetCheckH) 
    ! 
    ! write(*,*) "L743 Test image 77 saved:", trim(outSetCheck)!
    !--------------------------------

    ! smooth by three
    imImage3=-1.0!0.0
    do i=0,windowDX - 1
       do j=0,windowDY - 1
          nvsum=0.0
          do i1=i-1,i+1
             if (i1 .ge. 0 .and. i1 .lt. windowDX) then
                if (imImage2(i1,j) .ge. 0.0) then
                   imImage3(i,j)=imImage3(i,j)+imImage2(i1,j)
                   nvsum=nvsum+1.0
                endif
             endif
          enddo
          if (nvsum .gt. 1.0) imImage3(i,j)=imImage3(i,j)/nvsum
       enddo
    enddo

    !----------------------
    !write(*,*) "L771 creating testimage77.fit"
    !outSetCheck="testimage77.fit"
    !outDimCheck(1)=windowDX
    !outDimCheck(2)=windowDY
    !outDataSetCheckH = dataSet(trim(outSetCheck), Create)
    !outArrayCheckH=addArray(outDataSetCheckH, "PRIMARY", &
    !     real32, outDimCheck, "Counts", "IMAGE")!             
    !outOSWcheckP => real32Array2Data(outArrayCheckH)
    !outOSWcheckP = imImage3 
    !! Release handles on data.
    !call release(outArrayCheckH)        
    !call release(outDataSetCheckH) 
    ! 
    ! write(*,*) "L784 Test image 77 saved:", trim(outSetCheck)!
    !--------------------------------


    sliceX=0.0
    do i=0,windowDX - 1
       vsum=0.0
       nvsum=0.0
       do j=0,windowDY - 1
          if (imImage3(i,j) .ge. 0.0) then
             vsum=vsum+imImage3(i,j)
             nvsum=nvsum+1.0
          endif
       enddo
       if (nvsum .gt. 1.0) vsum=vsum/nvsum
       sliceX(i)=vsum
       !write(*,*) "L798 i=", i, " vsum=", vsum
    enddo
    !write(*,*) "***************************"

    sliceX2=0.0
    do i=0,8
       vsum=sliceX(i)
       i1=i+8
       if (i1 .lt. windowDX) vsum=vsum+sliceX(i1)
       i1=i+16
       if (i1 .lt. windowDX) vsum=vsum+sliceX(i1)
       sliceX2(i)=vsum
       !write(*,*) "L810 i=",i, " vsum=", vsum
    enddo


    ! compute the max and min values of the slice
    vMax=-1.0E9
    vMin=1.0E9
    nvMax=-1
    nvMin=-1
    do i=0,8
       if (sliceX2(i) .gt. vMax) then
          vMax=sliceX2(i)
          nvMax=i
       endif
       if (sliceX2(i) .lt. vMin) then
          vMin=sliceX2(i)
          nvMin=i
       endif
    enddo
    !write(*,*) "--------------------"
    !write(*,*) "L830 nvMax=", nvMax, " vMax=", vMax
    !write(*,*) "L831 nvMin=", nvMin, " vMin=", vMin
    maxMinDiff=abs(real(nvMax-nvMin))
    !write(*,*) "L833 maxMinDiff=", maxMinDiff
    inversedWeight=1.0+abs(maxMinDiff-4.0)
    !write(*,*) "L835 inversedWeight=", inversedWeight

    ! compute the 3-averaged values of vmax and vmin
    vMax=0.0
    vMin0=0.0
    do i=nvMax-1,nvMax+1
       i1=i
       if (i1 .lt. 0) i1=i1+9
       if (i1 .gt. 8) i1=i1-9
       vMax=vMax+sliceX2(i1)
       j1=i1+4
       if (j1 .lt. 0) j1=j1+9
       if (j1 .gt. 8) j1=j1-9
       vMin0=vMin0+sliceX2(j1)
    enddo
    !write(*,*) "L850 averaged  vMax=", vMax, " vMin0=", vMin0

    vMin=0.0
    do i=nvMin-1,nvMin+1
       i1=i
       if (i1 .lt. 0) i1=i1+9
       if (i1 .gt. 8) i1=i1-9
       vMin=vMin+sliceX2(i1)
    enddo
    !write(*,*) "L859 averaged vMin=", vMin
    vMaxmin=0.0
    if (vMax+vMin .gt. 0.0) vMaxmin=(vMax-vMin)/(vMax+vMin)
    if (vMax+vMin0 .gt. 0.0) vMaxmin0=(vMax-vMin0)/(vMax+vMin0)
    !write(*,*) "L863 vMax=", vMax, " vMin=", vMin, " vMin0=", vMin0
    !write(*,*) "L864 (max-min)/(max+min)=",vMaxmin  
    !write(*,*) "L838 vMaxmin0=", vMaxmin0
    mod8xfactor=vMaxmin/inversedWeight
    !write(*,*) "L867 mod8xfactor=", mod8xfactor

    !*******************************************
    ! repeat the mod-8 checking for the y-direction
    !*******************************************

    ! Subtract the cosecutive raws from one another
    imImage2=0.0
    ! up direction
    do j=0,windowDY - 2
       do i=0,windowDX - 1          
          imImage2(i,j)=imImage1(i,j)-imImage1(i,j+1)
       enddo
    enddo

    imImage4=0.0
    ! down direction
    do j=1,windowDY - 1
       do i=0,windowDX - 1
          imImage4(i,j)=imImage1(i,j)-imImage1(i,j-1)
       enddo
    enddo

    ! sum 
    ! and also remove all of the pixels within 6-pixel radius around the source
    do i=0,windowDX - 1
       do j=0,windowDY - 1
          imImage2(i,j)=imImage2(i,j)+imImage4(i,j)
          ! remove all of the pixels within 6-pixel radius around the source
          if (xsrc .gt. 0.0 .and. ysrc .gt. 0.0) then
             xpix=real(i)
             ypix=real(j)
             dist=(xpix-xsrc)*(xpix-xsrc)+(ypix-ysrc)*(ypix-ysrc)
             dist=sqrt(dist)
             if (dist .le. 6.0) imImage2(i,j)=-1.0!0.0
          endif
          !***************************************************
          ! 2011-09-29 nullify the extreme raws and columns
          !***************************************************
          if (i .eq. 0 .or. i .eq. windowDX-1 .or. j .eq. 0 .or. j .eq. windowDY-1) then
             imImage2(i,j)=-1.0!0.0
          endif
       enddo
    enddo

    ! smooth by three
    imImage3=-1.0!0.0
    do j=0,windowDY - 1
       do i=0,windowDX - 1
          nvsum=0.0
          do i1=j-1,j+1
             if (i1 .ge. 0 .and. i1 .lt. windowDY) then
                if (imImage2(i,i1) .ge. 0.0) then
                   imImage3(i,j)=imImage3(i,j)+imImage2(i,i1)
                   nvsum=nvsum+1.0
                endif
             endif
          enddo
          if (nvsum .gt. 1.0) imImage3(i,j)=imImage3(i,j)/nvsum
       enddo
    enddo

    sliceX=0.0
    do j=0,windowDY - 1
       vsum=0.0
       nvsum=0.0
       do i=0,windowDX - 1
          if (imImage3(i,j) .ge. 0.0) then
             vsum=vsum+imImage3(i,j)
             nvsum=nvsum+1.0
          endif
       enddo
       if (nvsum .gt. 1.0) vsum=vsum/nvsum
       sliceX(j)=vsum
       !write(*,*) "L943 j=", j, " vsum=", vsum
    enddo
    !write(*,*) "***************************"

    sliceX2=0.0
    do j=0,8
       vsum=sliceX(j)
       i1=j+8
       if (i1 .lt. windowDY) vsum=vsum+sliceX(i1)
       i1=i+16
       if (i1 .lt. windowDY) vsum=vsum+sliceX(i1)
       sliceX2(j)=vsum
       !write(*,*) "L958 i=",j, " vsum=", vsum
    enddo

    ! compute the max and min values of the slice
    vMax=-1.0E9
    vMin=1.0E9
    nvMax=-1
    nvMin=-1
    do j=0,8
       if (sliceX2(j) .gt. vMax) then
          vMax=sliceX2(j)
          nvMax=j
       endif
       if (sliceX2(j) .lt. vMin) then
          vMin=sliceX2(j)
          nvMin=j
       endif
    enddo

    !write(*,*) "--------------------"
    !write(*,*) "L929 nvMax=", nvMax, " vMax=", vMax
    !write(*,*) "L930 nvMin=", nvMin, " vMin=", vMin
    maxMinDiff=abs(real(nvMax-nvMin))
    !write(*,*) "L932 maxMinDiff=", maxMinDiff
    inversedWeight=1.0+abs(maxMinDiff-4.0)
    !write(*,*) "L934 inversedWeight=", inversedWeight

    ! compute the 3-averaged values of vmax and vmin
    vMax=0.0
    vMin0=0.0
    do j=nvMax-1,nvMax+1
       i1=j
       if (i1 .lt. 0) i1=i1+9
       if (i1 .gt. 8) i1=i1-9
       vMax=vMax+sliceX2(i1)
       j1=i1+4
       if (j1 .lt. 0) j1=j1+9
       if (j1 .gt. 8) j1=j1-9
       vMin0=vMin0+sliceX2(j1)
    enddo

    vMin=0.0
    do j=nvMin-1,nvMin+1
       i1=j
       if (i1 .lt. 0) i1=i1+9
       if (i1 .gt. 8) i1=i1-9
       vMin=vMin+sliceX2(i1)
    enddo

    vMaxmin=0.0
    if (vMax+vMin .gt. 0.0) vMaxmin=(vMax-vMin)/(vMax+vMin)
    if (vMax+vMin0 .gt. 0.0) vMaxmin0=(vMax-vMin0)/(vMax+vMin0)
    !write(*,*) "L961 vMax=", vMax, " vMin=", vMin, " vMin0=", vMin0
    !write(*,*) "L962 (max-min)/(max+min)=",vMaxmin  
    !write(*,*) "L963 vMaxmin0=", vMaxmin0
    mod8yfactor=vMaxmin/inversedWeight
    !write(*,*) "L965 mod8yfactor=", mod8yfactor


    deallocate (imImage1)
    deallocate (imImage2)
    deallocate (imImage3)
    deallocate (imImage4)
    deallocate (sliceX)
    deallocate (sliceX2)


end subroutine CheckMod8


subroutine CheckArrayBounds(array, i, j)
    real(kind=single), intent(in), dimension(: , :) :: array
    integer, intent(in) :: i, j
    integer :: iMax, jMax

    iMax = size(array, 1) - 1
    jMax = size(array, 2) - 1

    if (i .lt. 0) call fatal("RealArray bounds error- i < 0")
    if (j .lt. 0) call fatal("RealArray bounds error- j < 0")

    if (i .gt. iMax) then
       call fatal("RealArray bounds error- i > imax")
    end if

    if (j .gt. jMax) then
       call fatal("RealArray bounds error- j > jmax")
    end if

end subroutine CheckArrayBounds
subroutine CheckInt16ArrayBounds(array, i, j)
    integer(kind=int16), intent(in), dimension(: , :) :: array
    integer, intent(in) :: i, j
    integer :: iMax, jMax

    iMax = size(array, 1) - 1
    jMax = size(array, 2) - 1

    if (i .lt. 0) call fatal("Int16Array bounds error- i < 0")
    if (j .lt. 0) call fatal("Int16Array bounds error- j < 0")
    if (i .gt. iMax) call fatal("Int16Array bounds error- i > imax")
    if (j .gt. jMax) call fatal("Int16Array bounds error- j > jmax")

end subroutine CheckInt16ArrayBounds


end module omfastflat_mod






